from fastai.basics import os, np, pd, csv, torch, is_listy, delegates, typedispatch, Path, Tensor, TitledFloat, ColSplitter
from fastai.data.all import DataBlock, CategoryBlock, MultiCategoryBlock, DataLoaders, TransformBlock, ColReader, RandomSplitter
from fastai.vision.all import Image, TensorCategory, AddMaskCodes

from niftiai.core import plot, get_blended_image, TensorMask3d, TensorImage3d


def ImageBlock3d(cls=TensorImage3d):
    return TransformBlock(type_tfms=cls.create)


def MaskBlock3d(codes=None):
    return TransformBlock(type_tfms=TensorMask3d.create, item_tfms=AddMaskCodes(codes=codes))


class ImageDataLoaders3d(DataLoaders):
    @classmethod
    @delegates(DataLoaders.from_dblock)
    def from_df(cls, df, path='.', valid_pct=0.2, seed=None, fn_col=0, folder=None, suff='', label_col=1, label_delim=None,
                y_block=None, valid_col=None, item_tfms=None, batch_tfms=None, img_cls=TensorImage3d, bs=2, **kwargs):
        pref = f'{Path(path) if folder is None else Path(path)/folder}{os.path.sep}'
        if y_block is None:
            is_multi = (is_listy(label_col) and len(label_col) > 1) or label_delim is not None
            y_block = MultiCategoryBlock if is_multi else CategoryBlock
        splitter = RandomSplitter(valid_pct, seed=seed) if valid_col is None else ColSplitter(valid_col)
        dblock = DataBlock(blocks=(ImageBlock3d(img_cls), y_block),
                           get_x=ColReader(fn_col, pref=pref, suff=suff),
                           get_y=ColReader(label_col, label_delim=label_delim),
                           splitter=splitter,
                           item_tfms=item_tfms,
                           batch_tfms=batch_tfms)
        return cls.from_dblock(dblock, df, path=path, bs=bs, **kwargs)

    @classmethod
    def from_csv(cls, path, csv_fname='labels.csv', header='infer', delimiter=None, quoting=csv.QUOTE_MINIMAL, **kwargs):
        df = pd.read_csv(Path(path)/csv_fname, header=header, delimiter=delimiter, quoting=quoting)
        return cls.from_df(df, path=path, **kwargs)


@typedispatch
def show_batch(x: TensorImage3d, y: Tensor, samples, max_n=10, ctxs=None, **kwargs):
    title = [str(item) for item in samples.itemgot(1)]
    return x[:max_n].show(title=title, **kwargs)


@typedispatch
def show_results(x: TensorImage3d, y: Tensor, samples, outs, max_n=10, ctxs=None, **kwargs):
    title = []
    for item, pred, _ in zip(samples.itemgot(1), outs.itemgot(0), range(max_n)):
        item_str, pred_str = (str(pred)[1:6], str(item)[:5]) if isinstance(item, TitledFloat) else (str(item), str(pred))
        extra_str = '' if isinstance(item, TitledFloat) else [' ✖', ' ✔'][item == pred]
        title.append(f'Pred. {pred_str} is {item_str}' + extra_str)
    return x[:max_n].show(title=title, **kwargs)


@typedispatch
def plot_top_losses(x: TensorImage3d, y: TensorCategory, samples, outs, raws, losses, **kwargs):
    title = []
    for item, pred, r, loss in zip(samples.itemgot(1), outs.itemgot(0), raws, losses):
        title.append(f'Pred. {pred} is {item} (Loss={loss.item():.2g}, Prob.={r.max().item():.2g})')
    return x.show(title=title, **kwargs)


class SegmentationDataLoaders3d(ImageDataLoaders3d):
    @classmethod
    @delegates(DataLoaders.from_dblock)
    def from_df(cls, df, codes=None, y_block=None, **kwargs):
        y_block = MaskBlock3d(codes) if y_block is None else y_block
        return ImageDataLoaders3d.from_df(df, y_block=y_block, **kwargs)


@typedispatch
def show_batch(x: TensorImage3d, y: TensorMask3d, samples, ctxs=None, figsize=None, ctx=None, max_n=16, **kwargs):
    return plot(get_blended_image([x[:max_n, 0], y[:max_n]], **kwargs), figsize=figsize, ctx=ctx)


@typedispatch
def show_results(x: TensorImage3d, y: TensorMask3d, samples, outs, ctxs=None, max_n=16, figsize=None, ctx=None, nrows=None, **kwargs):
    nrows = len(x) if nrows is None else nrows
    targ_title, pred_title = (['Target'] + [None] * (nrows - 1), ['Prediction'] + [None] * (nrows - 1))
    im = get_results_image(x, y, samples, outs, max_n, nrows, targ_title, pred_title, **kwargs)
    return plot(im, figsize=figsize, ctx=ctx)


@typedispatch
def plot_top_losses(x: TensorImage3d, y: TensorMask3d, samples, outs, raws, losses, max_n=16, figsize=None, ctx=None, nrows=None, **kwargs):
    targ_title, pred_title = [], []
    for loss in losses:
        targ_title.append('Target')
        pred_title.append(f'Prediction loss={loss.item():.2g}')
    im = get_results_image(x, y, samples, outs, max_n, nrows, targ_title, pred_title, **kwargs)
    return plot(im, figsize=figsize, ctx=ctx)


def get_results_image(x, y, samples, outs, max_n, nrows, targ_title, pred_title, **kwargs):
    nrows = min(max_n, len(samples)) if nrows is None else nrows
    im0 = get_blended_image([x[:max_n, 0], y[:max_n]], nrows=nrows, title=targ_title, **kwargs)
    im1 = get_blended_image([x[:max_n, 0], torch.stack([*outs.itemgot(0)[:max_n]])], nrows=nrows, title=pred_title, **kwargs)
    return Image.fromarray(np.hstack([np.array(im0), np.array(im1)]))


if __name__ == '__main__':
    import glob
    import matplotlib.pyplot as plt
    from niftiai.transforms import Resize

    #fps = sorted(glob.glob('/home/lfisch/Data/mri/test_files_2/raw_nii/*.nii'))[:32]
    #fps_mask = sorted(glob.glob('/home/lfisch/Data/mri/test_files_2/brain_mask_nii/*.nii'))[:32]
    # fps = sorted(glob.glob('/mnt/data-ssd/neuro/T1w_images/test_files_gz/*.nii.gz'))[:32]
    # fps_mask = sorted(glob.glob('/mnt/data-ssd/neuro/T1w_images/masks_nii/*.nii'))[:32]
    fps = sorted(glob.glob('/home/lfisch/Projects/niftiai_beta/data/ds000001/*/anat/*.nii.gz'))
    fps_mask = sorted(glob.glob('/home/lfisch/Projects/niftiai_beta/data/ds000001/derivatives/deepmriprep/*/anat/mri/mask*.nii.gz'))
    df = pd.DataFrame({'filepath': fps, 'age': sum([], (len(fps) // 2) * ['a', 'b'])})

    # dls = ImageDataLoaders3d.from_df(df, path='/', fn_col='filepath', label_col='age', y_block=CategoryBlock(),
    #                                  item_tfms=[Resize(size=(128, 128, 128)),], bs=4)
    # dls.show_batch()

    #plt.show()

    df_ = pd.DataFrame({'filepath': fps, 'filepath_mask': fps_mask})
    dls = SegmentationDataLoaders3d.from_df(df_, path='/', fn_col='filepath', label_col='filepath_mask',
                                            item_tfms=[Resize(size=(128, 128, 128)),])
    dls.show_batch(cbar=True)#, use_affine=True)
    plt.show()



from fastai.basics import torch, random, store_attr, Iterable
from fastai.vision.augment import PadMode, RandTransform
from mriaug.core import (crop3d, affine3d, warp3d, affinewarp3d, chi_noise3d, bias_field3d, contrast,
                         ringing3d, motion3d, ghosting3d, spike3d, downsample3d, flip3d, dihedral3d)

from niftiai.core import TensorImage3d, TensorMask3d


class Crop3d(RandTransform):
    order = 40
    def __init__(self, size: (int, tuple), max_translate: (float, tuple) = 1.,
                 p: float = .5, batch: bool = False, item: bool = False):
        super().__init__(p=p)
        max_translate = tuple(3 * [max_translate]) if isinstance(max_translate, float) else max_translate
        size = (size, size, size) if isinstance(size, int) else tuple(size)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        bs = 1 if self.item else len(b[0])
        self.translate = get_rand_affine_param(bs, self.max_translate, device=b[0].device, batch=self.batch)

    def encodes(self, x: (TensorImage3d, TensorMask3d)):
        return crop3d(x, self.translate, self.size)


class AffineWarp3d(RandTransform):
    order = 40
    def __init__(self, max_zoom: (float, tuple) = .1, max_rotate: (float, tuple) = .1,
                 max_translate: (float, tuple) = .1, max_shear: (float, tuple) = .02, p_affine: float = .5,
                 max_warp: float = .01, k_size: (int, tuple) = 2, p_warp: float = .5, upsample: float = 2,
                 size: (int, tuple) = None, pad_mode: PadMode = PadMode.Zeros, batch: bool = False, item: bool = False):
        super().__init__()
        max_zoom = max_zoom if isinstance(max_zoom, Iterable) else tuple(3 * [max_zoom])
        max_rotate = max_rotate if isinstance(max_rotate, Iterable) else tuple(3 * [max_rotate])
        max_translate = max_translate if isinstance(max_translate, Iterable) else tuple(3 * [max_translate])
        max_shear = max_shear if isinstance(max_shear, Iterable) else tuple(3 * [max_shear])
        k_size = tuple(3 * [k_size]) if isinstance(k_size, int) else k_size
        size = None if size is None else (size, size, size) if isinstance(size, int) else tuple(size)
        store_attr()

    def before_call(self, b, split_idx):
        self.do_affine = random.random() < self.p_affine
        self.do_warp = random.random() < self.p_warp
        self.do = self.do_affine or self.do_warp
        bs = 1 if self.item else len(b[0])
        self.zoom = get_rand_affine_param(bs, self.max_zoom, device=b[0].device, batch=self.batch)
        self.rotate = get_rand_affine_param(bs, self.max_rotate, device=b[0].device, batch=self.batch)
        self.translate = get_rand_affine_param(bs, self.max_translate, device=b[0].device, batch=self.batch)
        self.shear = get_rand_affine_param(bs, self.max_shear, device=b[0].device, batch=self.batch)
        self.warp = self.max_warp * torch.rand(bs, device=b[0].device)
        self.k = torch.randn((bs, 3, *self.k_size), device=b[0].device)
        self.k[..., 0, 0, 0] = 0
        if self.batch and bs > 1:
            self.warp = self.warp[:1]
            self.k = self.k[:1]
        self.kws = {'do_affine': self.do_affine, 'do_warp': self.do_warp, 'zoom': self.zoom,
                    'rotate': self.rotate, 'translate': self.translate, 'shear': self.shear,
                    'warp': self.warp, 'k': self.k, 'upsample': self.upsample, 'size': self.size,
                    'mode': 'nearest' if isinstance(b[0], TensorMask3d) else 'bilinear', 'pad_mode': self.pad_mode}

    def encodes(self, x: TensorImage3d):
        return apply_affinewarp3d(x[None], **self.kws)[0] if self.item else apply_affinewarp3d(x, **self.kws)

    def encodes(self, x: TensorMask3d):
        return apply_affinewarp3d(x[None, None], **self.kws)[0, 0] if self.item else apply_affinewarp3d(x[:, None], **self.kws)[:, 0]


class Warp3d(RandTransform):
    order = 40
    def __init__(self, max_magnitude: float = .01, k_size: (int, tuple) = 2, p: float = .5, upsample: float = 2,
                 size: (int, tuple) = None, pad_mode: PadMode = PadMode.Zeros, batch: bool = False, item: bool = False):
        super().__init__(p=p)
        k_size = tuple(3 * [k_size]) if isinstance(k_size, int) else k_size
        size = None if size is None else (size, size, size) if isinstance(size, int) else tuple(size)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        bs = 1 if self.item else len(b[0])
        self.magnitude = self.max_magnitude * torch.rand(bs, device=b[0].device)
        self.k = torch.randn((bs, 3, *self.k_size), device=b[0].device)
        self.k[..., 0, 0, 0] = 0
        if self.batch and bs > 1:
            self.magnitude = self.magnitude[:1]
            self.k = self.k[:1]
        self.kws = {'magnitude': self.magnitude, 'k': self.k, 'upsample': self.upsample, 'size': self.size,
                    'mode': 'bilinear' if isinstance(b[0], TensorMask3d) else 'bilinear', 'pad_mode': self.pad_mode}

    def encodes(self, x: TensorImage3d):
        return warp3d(x[None], **self.kws)[0] if self.item else warp3d(x, **self.kws)

    def encodes(self, x: TensorMask3d):
        return warp3d(x[None, None], **self.kws)[0, 0] if self.item else warp3d(x[:, None], **self.kws)[:, 0]


class Affine3d(RandTransform):
    order = 40
    def __init__(self, max_zoom: (float, tuple) = .1, max_rotate: (float, tuple) = .1,
                 max_translate: (float, tuple) = .1, max_shear: (float, tuple) = .02, p: float = .5,
                 upsample: float = 2, size: (int, tuple) = None, pad_mode: PadMode = PadMode.Zeros,
                 batch: bool = False, item: bool = False):
        super().__init__(p=p)
        max_zoom = max_zoom if isinstance(max_zoom, Iterable) else tuple(3 * [max_zoom])
        max_rotate = max_rotate if isinstance(max_rotate, Iterable) else tuple(3 * [max_rotate])
        max_translate = max_translate if isinstance(max_translate, Iterable) else tuple(3 * [max_translate])
        max_shear = max_shear if isinstance(max_shear, Iterable) else tuple(3 * [max_shear])
        size = None if size is None else (size, size, size) if isinstance(size, int) else tuple(size)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        bs = 1 if self.item else len(b[0])
        self.zoom = get_rand_affine_param(bs, self.max_zoom, device=b[0].device, batch=self.batch)
        self.rotate = get_rand_affine_param(bs, self.max_rotate, device=b[0].device, batch=self.batch)
        self.translate = get_rand_affine_param(bs, self.max_translate, device=b[0].device, batch=self.batch)
        self.shear = get_rand_affine_param(bs, self.max_shear, device=b[0].device, batch=self.batch)
        self.kws = {'zoom': self.zoom, 'rotate': self.rotate, 'translate': self.translate, 'shear': self.shear,
                    'upsample': self.upsample, 'size': self.size, 'pad_mode': self.pad_mode,
                    'mode': 'nearest' if isinstance(b[0], TensorMask3d) else 'bilinear'}

    def encodes(self, x: TensorImage3d):
        return affine3d(x[None], **self.kws)[0] if self.item else affine3d(x, **self.kws)

    def encodes(self, x: TensorMask3d):
        return affine3d(x[None, None], **self.kws)[0, 0] if self.item else affine3d(x[:, None], **self.kws)[:, 0]


class Zoom3d(Affine3d):
    def __init__(self, max_zoom: (float, tuple) = .1, p: float = .5, upsample: float = 2,
                 size: (int, tuple) = None, pad_mode: PadMode = PadMode.Zeros, batch: bool = False, item: bool = False):
        super().__init__(max_zoom=max_zoom, max_rotate=0, max_translate=0, max_shear=0, p=p,
                         upsample=upsample, size=size, pad_mode=pad_mode, batch=batch, item=item)


class Rotate3d(Affine3d):
    def __init__(self, max_rotate: (float, tuple) = .1, p: float = .5, upsample: float = 2,
                 size: (int, tuple) = None, pad_mode: PadMode = PadMode.Zeros, batch: bool = False, item: bool = False):
        super().__init__(max_zoom=0, max_rotate=max_rotate, max_translate=0, max_shear=0, p=p,
                         upsample=upsample, size=size, pad_mode=pad_mode, batch=batch, item=item)


class Translate3d(Affine3d):
    def __init__(self, max_translate: (float, tuple) = .1, p: float = .5, upsample: float = 2.,
                 size: (int, tuple) = None, pad_mode: PadMode = PadMode.Zeros, batch: bool = False, item: bool = False):
        super().__init__(max_zoom=0, max_rotate=0, max_translate=max_translate, max_shear=0, p=p,
                         upsample=upsample, size=size, pad_mode=pad_mode, batch=batch, item=item)


class Shear3d(Affine3d):
    def __init__(self, max_shear: (float, tuple) = .02, p: float = .5, upsample: float = 2,
                 size: (int, tuple) = None, pad_mode: PadMode = PadMode.Zeros, batch: bool = False, item: bool = False):
        super().__init__(max_zoom=0, max_rotate=0, max_translate=0, max_shear=max_shear, p=p,
                         upsample=upsample, size=size, pad_mode=pad_mode, batch=batch, item=item)


class ChiNoise3d(RandTransform):
    order = 50
    def __init__(self, max_intensity: float = .1, max_dof: int = 3, p: float = .5, batch: bool = False):
        super().__init__(p=p)
        store_attr()

    def encodes(self, x: TensorImage3d):
        return chi_noise3d(x, self.max_intensity, random.randint(1, self.max_dof), self.batch)


class BiasField3d(RandTransform):
    order = 55
    def __init__(self, max_intensity: float = .1, k_size: (int, tuple) = 2, p: float = .5, batch: bool = False):
        super().__init__(p=p)
        k_size = tuple(3 * [k_size]) if isinstance(k_size, int) else k_size
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.intensity = self.max_intensity * torch.randn(len(b[0]), device=b[0].device)
        self.k = torch.randn((*b[0].shape[:-3], *self.k_size), device=b[0].device)
        self.k[..., 0, 0, 0] = 0
        if self.batch:
            self.intensity = self.intensity[:1]
            self.k = self.k[:1]

    def encodes(self, x: TensorImage3d):
        return bias_field3d(x, self.intensity, k=self.k)


class Contrast3d(RandTransform):
    order = 60
    def __init__(self, max_lighting: float = .2, p: float = .5, batch: bool = False):
        super().__init__(p=p)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.lighting = self.max_lighting * torch.rand(len(b[0]), dtype=b[0].dtype, device=b[0].device)
        if self.batch:
            self.lighting = self.lighting[:1]

    def encodes(self, x: TensorImage3d):
        return contrast(x, self.lighting)


class Ringing3d(RandTransform):
    order = 65
    def __init__(self, max_intensity: float = .2, frequency: float = .7,
                 dims: (int, tuple) = 2, p: float = .5, batch: bool = False):
        super().__init__(p=p)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.intensity = self.max_intensity * torch.rand(len(b[0]), dtype=b[0].dtype, device=b[0].device)
        self.dim = self.dims if isinstance(self.dims, int) else self.dims[-random.randint(0, len(self.dims) - 1)]
        if self.batch and len(b[0]) > 1:
            self.intensity = self.intensity[:1]

    def encodes(self, x: TensorImage3d):
        return ringing3d(x, self.intensity, self.frequency, self.dim)


class Motion3d(RandTransform):
    order = 70
    def __init__(self, max_intensity: float = .2, max_move: float = .02,
                 p: float = .5, batch: bool = False, item: bool = False):
        super().__init__(p=p)
        max_move = tuple(3 * [max_move]) if isinstance(max_move, float) else max_move
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        bs = 1 if self.item else len(b[0])
        self.intensity = self.max_intensity * torch.rand(bs, dtype=b[0].dtype, device=b[0].device)
        self.move = get_rand_affine_param(bs, self.max_move, b[0].device, self.batch)
        if self.batch and len(b[0]) > 1:
            self.intensity = self.intensity[:1].repeat(bs)

    def encodes(self, x: TensorImage3d):
        return motion3d(x[None], self.intensity, self.move)[0] if self.item else motion3d(x, self.intensity, self.move)


class Ghosting3d(RandTransform):
    order = 75
    def __init__(self, max_intensity: float = .2, n_ghosts: int = 4,
                 dims: (int, tuple) = 2, p: float = .5, batch: bool = False):
        super().__init__(p=p)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.intensity = self.max_intensity * torch.rand(len(b[0]), dtype=b[0].dtype, device=b[0].device)
        self.dim = self.dims if isinstance(self.dims, int) else self.dims[-random.randint(0, len(self.dims) - 1)]
        if self.batch and len(b[0]) > 1:
            self.intensity = self.intensity[:1]

    def encodes(self, x: TensorImage3d):
        return ghosting3d(x, self.intensity, self.n_ghosts, self.dim)


class Spike3d(RandTransform):
    order = 80
    def __init__(self, max_intensity: float = .5, max_frequency: float = 1., p: float = .5, batch: bool = False):
        super().__init__(p=p)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.intensity = self.max_intensity * torch.rand(len(b[0]), dtype=b[0].dtype, device=b[0].device)
        self.frequencies = self.max_frequency * torch.rand((len(b[0]), 3), dtype=b[0].dtype, device=b[0].device)
        if self.batch and len(b[0]) > 1:
            self.intensity = self.intensity[:1].repeat(len(b[0]))
            self.frequencies = self.frequencies[:1].repeat(len(b[0]), 1)

    def encodes(self, x: TensorImage3d):
        return spike3d(x, self.intensity, self.frequencies)


class Downsample3d(RandTransform):
    order = 85
    def __init__(self, max_downsample: float = 2., dims: (int, tuple) = 2,
                 p: float = .5, batch: bool = False, item: bool = False):
        super().__init__(p=p)
        store_attr()

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        bs = 1 if self.item else len(b[0])
        self.scale = 1 / (1 + self.max_downsample * torch.rand(bs, device=b[0].device))
        self.dim = self.dims if isinstance(self.dims, int) else random.choice(self.dims)

    def encodes(self, x: TensorImage3d):
        kws = {'scale': self.scale, 'dim': self.dim, 'mode': 'nearest'}
        return downsample3d(x[None], **kws)[0] if self.item else downsample3d(x, **kws)


class Flip3d(RandTransform):
    order = 90
    def __init__(self, p: float = .5):
        super().__init__(p=p)

    def encodes(self, x: (TensorImage3d, TensorMask3d)):
        return flip3d(x)


class Dihedral3d(RandTransform):
    order = 95
    def __init__(self, p: float = .5, ks: tuple = None):
        super().__init__(p=p)
        store_attr()
        self.ks = tuple(range(24)) if ks is None else ks

    def encodes(self, x: (TensorImage3d, TensorMask3d)):
        return dihedral3d(x, k=random.choice(self.ks))


def apply_affinewarp3d(x, do_affine, do_warp, zoom, rotate, translate, shear, warp, k, upsample, size, mode, pad_mode):
    if do_affine and do_warp:
        return affinewarp3d(x, zoom=zoom, rotate=rotate, translate=translate, shear=shear, magnitude=warp,
                            k=k, size=size, mode=mode, upsample=upsample, pad_mode=pad_mode)
    elif do_warp:
        return warp3d(x, magnitude=warp, k=k, size=size, mode=mode, upsample=upsample, pad_mode=pad_mode)
    else:
        return affine3d(x, zoom=zoom, rotate=rotate, translate=translate, shear=shear,
                        size=size, mode=mode, upsample=upsample, pad_mode=pad_mode)


def get_rand_affine_param(bs, max_param, device, batch):
    param = (2 * torch.rand((bs, 3), device=device) - 1) * torch.tensor(max_param, device=device)
    return param[:1].repeat(bs, 1) if batch and bs > 1 else param


def aug_transforms3d(mult: float = 1., max_warp: float = .01, warp_k_size: int = 2, p_warp: float = .2,
                     max_zoom: (float, tuple) = .1, max_rotate: (float, tuple) = .1,
                     max_translate: (float, tuple) = .1, max_shear: float = .02, p_affine: float = .5,
                     max_noise: float = .1, max_dof_noise: int = 4, p_noise: float = .1, max_bias: float = .1,
                     bias_k_size: int = 2, p_bias: float = .1, max_contrast: float = .2, p_contrast: float = .1,
                     max_ring: float = .2, freq_ring: float = .7, dims_ring: (int, tuple) = 2, p_ring: float = .1,
                     max_motion: float = .2, max_move: float = .02, p_motion: float = .1, max_ghost: float = .2,
                     n_ghosts: int = 4, dims_ghost: (int, tuple) = 2, p_ghost: float = .1, max_spike: float = .1,
                     freq_spike: float = 1., p_spike: float = .01, max_down: float = 2., dims_down: (int, tuple) = 2,
                     p_down: float = .1, p_flip: float = .5, p_dihedral: float = .0, batch: bool = False,
                     item: bool = False, **kwargs):
    do_zoom = all(p != 0 for p in max_zoom) if isinstance(max_zoom, Iterable) else max_zoom != 0
    do_rotate = all(p != 0 for p in max_rotate) if isinstance(max_rotate, Iterable) else max_rotate != 0
    do_translate = all(p != 0 for p in max_translate) if isinstance(max_translate, Iterable) else max_translate != 0
    do_shear = all(p != 0 for p in max_shear) if isinstance(max_shear, Iterable) else max_shear != 0
    mult_max_zoom = tuple(mult * p for p in max_zoom) if isinstance(max_zoom, Iterable) else mult * max_zoom
    mult_max_rotate = tuple(mult * p for p in max_rotate) if isinstance(max_rotate, Iterable) else mult * max_rotate
    mult_max_translate = tuple(mult * p for p in max_translate) if isinstance(max_translate, Iterable) else mult * max_translate
    mult_max_shear = tuple(mult * p for p in max_shear) if isinstance(max_shear, Iterable) else mult * max_shear
    res = []
    if max_warp:
        if do_zoom or do_rotate or do_translate or do_shear:
            res.append(AffineWarp3d(mult_max_zoom, mult_max_rotate, mult_max_translate, mult_max_shear,
                                    p_affine=p_affine, max_warp=mult * max_warp, k_size=warp_k_size,
                                    p_warp=p_warp, batch=batch, item=item, **kwargs))
        else:
            res.append(Warp3d(mult * max_warp, k_size=warp_k_size, p=p_warp, batch=batch, item=item))
    else:
        if do_translate or do_rotate or do_zoom or do_shear:
            res.append(Affine3d(mult_max_zoom, mult_max_rotate, mult_max_translate, mult_max_shear,
                                p=p_affine, batch=batch, item=item, **kwargs))
    if max_noise: res.append(ChiNoise3d(mult * max_noise, max_dof=max_dof_noise, p=p_noise, batch=batch))
    if max_bias: res.append(BiasField3d(mult * max_bias, k_size=bias_k_size, p=p_bias, batch=batch))
    if max_contrast: res.append(Contrast3d(mult * max_contrast, p=p_contrast, batch=batch))
    if max_ring: res.append(Ringing3d(mult * max_ring, freq_ring, dims=dims_ring, p=p_ring, batch=batch))
    if max_motion: res.append(Motion3d(mult * max_motion, max_move=max_move, p=p_motion, batch=batch, item=item))
    if max_ghost: res.append(Ghosting3d(mult * max_ghost, n_ghosts=n_ghosts, dims=dims_ghost, p=p_ghost, batch=batch))
    if max_spike: res.append(Spike3d(mult * max_spike, max_frequency=freq_spike, p=p_spike, batch=batch))
    if max_down: res.append(Downsample3d(mult * max_down, dims=dims_down, p=p_down, batch=batch, item=item))
    if p_flip: res.append(Flip3d(p=p_flip))
    if p_dihedral: res.append(Dihedral3d(p=p_dihedral))
    return res



import nibabel as nib
from niftiview import NiftiImageGrid
from niftiview.image import CMAP, QRANGE
from niftiview.core import load_np, load_nib
from fastai.basics import np, plt, torch, ndarray, warnings, Path, Tensor, TensorImageBW


class TensorImageBase3d(TensorImageBW):
    _show_args = {'cmap': CMAP, 'transp_if': None, 'qrange': QRANGE[0], 'resizing': 1}

    def show(self, use_affine=None, figsize=None, ctx=None, **kwargs):
        im = self.get_image(use_affine=use_affine, **kwargs)
        return plot(im, figsize=figsize, ctx=ctx)

    def get_image(self, use_affine=True, **kwargs):
        if use_affine:
            use_affine = self.org_shape[-3:] == self.shape[-3:] if hasattr(self, 'org_shape') else False
        for k, v in self._show_args.items():
            if k not in kwargs:
                kwargs.update({k: v})
        arrays = channels_last(self.detach().cpu().numpy())
        arrays = [arrays] if self.ndim < 5 else arrays
        affines = len(self) * [self.affine if hasattr(self, 'affine') and use_affine else None]
        return NiftiImageGrid(arrays=arrays, affines=affines).get_image(**kwargs)

    def save(self, fn: (str, Path), squeeze=True, no_warning=False, cls=nib.Nifti1Image, **kwargs):
        array = self.detach().cpu().numpy()
        if str(fn).endswith('.npy'):
            np.save(fn, array, **kwargs)
        else:
            array = channels_last(array)
            if squeeze and array.ndim > 3:
                array = array.squeeze()
            if array.ndim > 4 and not no_warning:
                warnings.warn(f'Could not squeeze array into 4D. Saved {array.ndim}D array', UserWarning)
            array = orient_array(array, self.affine)
            cls(array, affine=self.affine, header=self.header).to_filename(fn, **kwargs)

    @classmethod
    def create(cls, fn: (str, Path, ndarray, Tensor), affine=None, header=None, filepath=None, **kwargs):
        if isinstance(fn, (str, Path)):
            filepath = str(fn)
            fn, affine, header = load_np(filepath) if filepath.endswith('.npy') else load_nib(filepath)
            fn = tensor_reshape(fn, is_mask='TensorMask3d' in str(cls))
        if isinstance(fn, ndarray):
            fn = torch.from_numpy(fn)
        return cls.__new__(cls=cls, x=fn, affine=affine, header=header, filepath=filepath, org_shape=fn.shape, **kwargs)


class TensorImage3d(TensorImageBase3d):
    pass


class TensorMask3d(TensorImageBase3d):
    _show_args = {'cmap': 'colorbrewer:reds', 'transp_if': '=0', 'qrange': QRANGE[1], 'resizing': 0}

    def show(self, **kwargs):
        codes = getattr(self, 'codes', None)
        if codes is not None:
            vmax = max(codes)
            kwargs['vrange'] = (0, vmax)
            kwargs['cmap'] = 'colorbrewer:' + ('reds' if vmax < 3 else f'set1_{vmax}_r' if vmax < 10 else 'set1_r')
        return super().show(**kwargs)
        # return TensorImage3D(self).show(**kwargs)


def plot(im, ax=None, figsize=None, ctx=None):
    ax = ctx if ax is None else ctx
    if figsize is None:
        figsize = (im.size[0] / 100, im.size[1] / 100)
    if ax is None:
        _, ax = plt.subplots(figsize=figsize, constrained_layout=True)
    ax.imshow(im.convert('RGBA'), cmap=None)
    ax.axis('off')
    return ax


def tensor_reshape(x, is_mask=False):
    if x.ndim == 3  and is_mask:
        return x#.astype(np.uint8)
    x = x[..., None] if x.ndim == 3 else x
    return channels_second(x)


def get_blended_image(xs, use_affine=False, **kwargs):
    use_affine = use_affine and all(x.org_shape[-3:] == x.shape[-3:] for x in xs)
    for k in TensorImageBase3d._show_args:
        kwargs.update({k: kwargs[k] if k in kwargs else [x._show_args[k] for x in xs]})
    arrays, affines = [], []
    for i in range(len(xs[0])):
        arrays.append([x[i].detach().cpu().numpy() for x in xs])
        affines.append([x[i].affine if use_affine else None for x in xs])
    return NiftiImageGrid(arrays=arrays, affines=affines).get_image(**kwargs)


def orient_array(array, affine):
    ornt = nib.orientations.ornt_transform(start_ornt=[[0, 1], [1, 1], [2, 1]], end_ornt=nib.io_orientation(affine))
    return nib.apply_orientation(array, ornt)


def channels_last(array):
    return array if array.ndim < 4 else array.transpose(*list(range(array.ndim - 4)), -3, -2, -1, -4)


def channels_second(array):
    return array if array.ndim < 4 else array.transpose(*list(range(array.ndim - 4)), -1, -4, -3, -2)



from .augment import aug_transforms3d
from .core import TensorImage3d, TensorMask3d
from .data import ImageDataLoaders3d, SegmentationDataLoaders3d
from .learner import cnn_learner3d, unet_learner3d
from .transforms import Scale



from fastai.basics import delegates, Learner, DiceLoss

from .models import ResNet3d, UNet3d, LinkNet3d


@delegates(Learner.__init__)
def cnn_learner3d(dls, model=None, c_in=None, n_out=None, **kwargs):
    if model is None:
        n_in = dls.one_batch()[0].shape[1] if c_in is None else c_in
        n_out = dls.c if n_out is None else n_out
        assert n_out, '`n_out` is not defined, and could not be inferred from data, set `dls.c` or pass `n_out`'
        img_size = dls.one_batch()[0].shape[-3:]
        model = ResNet3d(shape=img_size, n_out=n_out, c_in=n_in)
    return Learner(dls, model, **kwargs)


@delegates(Learner.__init__)
def unet_learner3d(dls, c_out, linknet=False, model=None, c_in=None, loss_func=None, **kwargs):
    if model is None:
        c_in = dls.one_batch()[0].shape[1] if c_in is None else c_in
        model = LinkNet3d(c_in=c_in, c_out=c_out) if linknet else UNet3d(c_in=c_in, c_out=c_out)
    loss_func = DiceLoss() if loss_func is None else loss_func
    return Learner(dls=dls, model=model, loss_func=loss_func, **kwargs)



import torch
import torch.nn as nn
NORM_CLS = nn.BatchNorm3d
ACT_MODULE = nn.ReLU(inplace=True)


class ResNet3d(nn.Sequential):
    def __init__(self, shape, c_in=1, n_out=1, c_start=16, n_blocks=(3,), kernel=3, pool1_kernel=2,
                 p_dropout=.5, cat_pool=False, bias=False, norm_cls=NORM_CLS, act_module=ACT_MODULE):
        conv1 = nn.Conv3d(c_in, c_start, kernel, stride=2, bias=bias)
        pool1 = nn.MaxPool3d(kernel_size=pool1_kernel, stride=2, padding=1)
        norm1 = norm_cls(c_start)
        res_blocks = []
        for i, nb in enumerate(n_blocks):
            stride = 1 if i in [0, len(n_blocks) - 1] else [2 if e > max(shape) // 2 else 1 for e in shape]
            c_in = 2 ** i * c_start
            block1 = ResBlock(c_in, c_in * 2, stride=stride, bias=bias, norm_cls=norm_cls, act_module=act_module)
            blocks = nb * [ResBlock(c_in * 2, c_in * 2, stride=1, bias=bias, norm_cls=norm_cls, act_module=act_module)]
            res_blocks.append(nn.Sequential(block1, *blocks))
        res_blocks = nn.Sequential(*res_blocks)
        final_pool = AdaptiveConcatPool3d() if cat_pool else nn.AdaptiveAvgPool3d(1)
        n_flatten = 2 ** len(n_blocks) * c_start * 2 if cat_pool else 2 ** len(n_blocks) * c_start
        head = ClassifierHead(nodes=[n_flatten, n_out], drops=[p_dropout], act_module=act_module)
        super().__init__(conv1, pool1, nn.ReLU(), norm1, res_blocks, final_pool, nn.Flatten(), head)


class ResBlock(nn.Module):
    def __init__(self, c_in, c_out, stride=1, bias=False, norm_cls=NORM_CLS, act_module=ACT_MODULE):
        super().__init__()
        self.convs = nn.Sequential(nn.Conv3d(c_in, c_out, 3, padding=1, stride=stride, bias=bias),
                                   act_module, norm_cls(c_out),
                                   nn.Conv3d(c_out, c_out, 3, padding=1, bias=bias), norm_cls(c_out))
        if stride > 1 or (c_in != c_out):
            self.down = nn.Sequential(nn.Conv3d(c_in, c_out, 1, stride=stride, bias=bias), norm_cls(c_out))
        else:
            self.down = None
        self.act = act_module

    def forward(self, x):
        res = x if self.down is None else self.down(x)
        return self.act(self.convs(x) + res)


class ClassifierHead(nn.Sequential):
    def __init__(self, nodes, drops, act_module=ACT_MODULE):
        modules = []
        activs = [act_module] * (len(nodes) - 2) + [None]
        for n_in, n_out, p, actn in zip(nodes[:-1], nodes[1:], drops, activs):
                modules.append(nn.BatchNorm1d(n_in))
                if p != 0:
                    modules.append(nn.Dropout(p))
                modules.append(nn.Linear(n_in, n_out))
                if actn is not None:
                    modules.append(actn)
        super().__init__(*modules)


class AdaptiveConcatPool3d(nn.Module):
    def __init__(self, sz=None):
        super().__init__()
        sz = sz or 1
        self.ap, self.mp = nn.AdaptiveAvgPool3d(sz), nn.AdaptiveMaxPool3d(sz)

    def forward(self, x):
        return torch.cat([self.mp(x), self.ap(x)], dim=1)



from resize_right import resize, interp_methods
from fastai.basics import np, noop, torch, to_cpu, store_attr, Normalize, DisplayedTransform

from .core import TensorMask3d, TensorImage3d


class Crop(DisplayedTransform):
    order = 10

    def __init__(self, slices, **kwargs):
        store_attr()
        super().__init__(**kwargs)

    def encodes(self, x: (TensorMask3d, TensorImage3d)):
        return x[self.slices]


class ReplaceNan(DisplayedTransform):
    order = 20
    def __init__(self, nan: float = 0., **kwargs):
        store_attr()
        super().__init__(**kwargs)

    def encodes(self, x: TensorImage3d):
        return torch.nan_to_num(x, nan=self.nan)


class Scale(DisplayedTransform):
    order = 30
    def __init__(self, maximum: float = .99, minimum: float = .0, are_values: bool = False, **kwargs):
        store_attr()
        super().__init__(**kwargs)

    def encodes(self, x: TensorImage3d):
        if self.are_values:
            return (x - self.minimum) / (self.maximum - self.minimum)
        else:
            minimum, maximum = np.quantile(x.cpu().numpy(), [self.minimum, self.maximum])
            return (x - minimum) / (maximum - minimum)


class Resize(DisplayedTransform):
    order = 40

    def __init__(self, size: (int, tuple), pad_mode: str = 'constant', antialiasing: bool = True, **kwargs):
        size = (size, size, size) if isinstance(size, int) else size
        store_attr()
        super().__init__(**kwargs)

    def encodes(self, x: TensorImage3d):
        return resize(x, out_shape=[*x.shape[:-3], *self.size], pad_mode=self.pad_mode,
                      antialiasing=self.antialiasing, interp_method=interp_methods.lanczos3, by_convs=True)

    def encodes(self, x: TensorMask3d):
        return resize(x, out_shape=[*x.shape[:-3], *self.size], pad_mode=self.pad_mode,
                      antialiasing=False, interp_method=interp_methods.linear, by_convs=True)


class Normalize3D(Normalize):
    def __init__(self, mean=None, std=None, axes=(0, 2, 3, 4)):
        super().__init__(mean=mean, std=std, axes=axes)

    def encodes(self, x:TensorImage3d):
        return (x-self.mean) / self.std

    def decodes(self, x:TensorImage3d):
        f = to_cpu if x.device.type=='cpu' else noop
        return (x*f(self.std) + f(self.mean))
